; Hardware functions for Flash-ROM loader
; Written in Assembler for speed (needed for Atmel)

; D-X Designs Pty. Ltd.   November 1995

; HARDWARE.ASM

;Timing requirement for Atmel: no more than 150uS between bytes in a
; block-load. Using 64-byte blocks, this gives 9.6mS/block.
; Block-write routine timed (including loop o'heads) at 1563 loops in
; 10 secs. [1E6 bytes written], or loop-time of 6.4mS (60% of limit), 
; on a 486-DX/50. So write time is approx. 100uS / byte.

;Address format:
; Chip designator: 0 = ROM, 1 = RAM
; Address bits: A[16:0] - MS first

    .MODEL   compact,C
    .CODE
    .386
    PUBLIC   C ReadBlock,  WriteBlock, OneStep, SetAddress
    PUBLIC   C BumpAddr,   WriteByte,  ReadByte
    EXTRN    C LPort:word, chipsel:dword

adlen	equ	18		;Length of address field (incl. chip select)

cycle macro    ;;Clock the TAP, sample data
      push     bx
      mov      bl,al
      call     getinp             ;;Sample before clock
      xchg     bl,al
      call     pulseclock         ;;Pulse the TAP clock
      mov      al,bl
      pop      bx                 ;;A[6] = TDO, A[7] = DONE
      endm

;====================================================================

; Read data from PROM to program buffer

;   void     ReadBlock (        //No returns
; 	     byte far *LPdata,  //Buffer address
;            word Count,        //Buffer length
;            long PROMaddr)     //Starting PROM address

ReadBlock    PROC  C LPdata:DWORD, Count:WORD, PROMaddr:DWORD
    push     si
    push     di
    mov      dx,LPort
    mov      ebx,dword ptr PROMaddr ;NB PROM addresses are 18-bits
    call     setaddress             ;Load LCA address counter (from EBX)
    les      si,LPdata
    mov      di,Count

rloop:
    call     readbyte           ;Get 1 byte from LCA (to AL)
    mov      es:[si],al
    inc      si                 ;Bump pointer
    call     bumpaddr           ;Bump hardware address
    dec      di
    jnz      rloop              ;Leave CX free for fast inner loops

    pop      di                 ;Then clean-up & quit
    pop      si
    ret
ReadBlock    ENDP

;====================================================================

; Write data from program buffer to PROM

;   void     WriteBlock (       //No returns
;            byte far *LPdata,  //Buffer address
;            word Count,        //Buffer length
;            long PROMaddr)     //Starting PROM address

WriteBlock   PROC  C LPdata:DWORD, Count:WORD, PROMaddr:DWORD
    push     si
    push     di
    mov      dx,LPort
    
    mov      ebx,dword ptr PROMaddr
;    mov      di,word ptr PROMaddr+2
    call     setaddress         ;Load LCA address counter (from EBX)
    
    les      si,LPdata
    mov      di,Count

wloop:
    mov      al,es:[si]         ;Get byte from buffer
    inc      si                 ;Bump pointer
    call     writebyte          ;Put byte to LCA (from AL)
    call     bumpaddr           ;Bump hardware address
    dec      di
    jnz      wloop              ;Leave CX free for fast inner loops

    pop      di                 ;Then clean-up & quit
    pop      si
    ret
WriteBlock   ENDP

;====================================================================

; Low-level functions: set hardware PROM pointer

;   void     SetAddress (       //No returns
;            long PROMaddr)     //18-bit PROM address to set

SetAddress   PROC  C PROMaddr:DWORD
    push     di
    mov      dx,LPort
    mov      ebx,dword ptr PROMaddr ;NB PROM addresses are 18-bits
    call     setaddress             ;Load LCA address counter (from EBX)
    pop      di
    ret
SetAddress   ENDP

;====================================================================

; Low-level functions: increment the PROM address pointer

;   void BumpAddr ()            //No returns

BumpAddr     PROC C
    mov      dx,LPort
    call     bumpaddr
    ret
BumpAddr     ENDP

;====================================================================

; Low-level functions: write 1 byte to PROM

;   void     WriteByte (        //No returns
;            byte data )        //To write

WriteByte    PROC  C data:WORD
    mov      dx,LPort
    mov      ax,data            ;Just AL that we want
    call     writebyte
    ret
WriteByte    ENDP

;====================================================================

; Low-level functions: read 1 byte from PROM

;    byte    ReadByte ()        //Return byte as read

ReadByte     PROC  C
    mov      dx,LPort
    call     readbyte
    ret
ReadByte     ENDP

;====================================================================

; Low-level functions: run one step of the TAP port

;   byte     OneStep (          //Returns bit 0 = DONE, bit 7 = TDO
;            TMS,               //TRUE for TMS signal
;            TDI)               //TRUE for TDI signal


TMS          equ 10H            ;Bit values for TMS & TDI
TDI          equ 20H            ; Values for high signals

OneStep      PROC  C tms:WORD, tdi:WORD
    mov      al,0
    mov      cx,tms
    jcxz     os1
    or       al,TMS             ;Set up the bits to send
os1:
    mov      cx,tdi
    jcxz     os2
    or       al,TDI
os2:
    mov      dx,LPort
    cycle
    rol      al,1               ;AL[0] = DONE, AL[7] = TDO
    ret
OneStep      ENDP

;====================================================================

setaddress   PROC near          ;Load the LCA address counter (EBX)
    mov      cx,adlen-1         ;Bit count less 1
    add      ebx,chipsel        ;Modifier for chip socket
    shl      ebx,32-adlen	;Shift up to the top

    mov      al,0
    call     pulseclock
    mov      al,TMS
    call     pulseclock         ;Sent the command
    mov      al,TDI
    call     pulseclock
slp:
    mov      al,0
    shl      ebx,1              ;Bits => CY, MS first
    jnc      s2
    mov      al,TDI             ;TMS low, TDI = data
s2:
    call     pulseclock
    loop     slp                ;First 17 bits (MS end)
    mov      al,TMS
    shl      ebx,1              ;Final/LS bit (TMS set, here)
    jnc      s3
    mov      al,TMS + TDI       ;TMS & TDI set up
s3:
    call     pulseclock         ;Sent final data bit
    ret
setaddress   ENDP

;====================================================================

bumpaddr     PROC near          ;Increment the address counter
    mov      al,0
    call     pulseclock
    mov      al,TMS + TDI
    call     pulseclock
    mov      al,TMS + TDI
    call     pulseclock
    ret
bumpaddr     ENDP

;====================================================================

readbyte     PROC near          ;Get a byte to AL
    mov      al,TDI
    call     pulseclock         ;Send the command
    mov      al,TMS + TDI
    call     pulseclock
    mov      al,TMS + TDI       ;Trigger 1st. bit
    call     pulseclock
    call     getinp             ;NB Reverse of usual "cycle" order
    rcl      al,2               ;Bit in CY
    rcl      ah,1               ;Put into AH
    mov      cx,7
rlp:
    mov      al,TDI
    call     pulseclock
    call     getinp
    rcl      al,2
    rcl      ah,1
    loop     rlp
    mov      al,TDI + TMS
    call     pulseclock         ;Switch back to idle
    mov      al,ah              ;Byte to return
    ret
readbyte     ENDP

;====================================================================

writebyte    PROC near          ;Write a byte from AL
    mov      ah,al              ;Data in AH
    mov      cx,7               ;Bit count
    mov      al,TDI
    call     pulseclock
    mov      al,TMS
    call     pulseclock
wlp:
    mov      al,0
    shl      ah,1               ;Bits => CY, MS first
    jnc      w2
    mov      al,TDI             ;TMS low, TDI = data
w2:
    call     pulseclock
    loop     wlp                ;First 7 bits
    mov      al,TMS
    shl      ah,1               ;Final bit (TMS set, here)
    jnc      w3
    mov      al,TMS + TDI       ;TMS & TDI set up
w3:
    call     pulseclock         ;Sent final data bit
    mov      al,TMS + TDI
    call     pulseclock         ;Final bits (write pulse)
    mov      al,TMS + TDI
    call     pulseclock
    ret
writebyte    ENDP

;====================================================================

DPidle       equ 0BH            ;Data-port idle state
DPclkset     equ 02H            ;Sets TCK high
DPclkrst     equ 01H            ;Sets TCK low

pulseclock   PROC near          ;Given TMS & TDI, hit the clock
    or       al,DPidle          ;Set up the other bits
    out      (dx),al
    and      al, not DPclkset
    out      (dx),al            ;TCK sets high
    or       al,DPclkset
    out      (dx),al            ;Clock high-period (settles)
    and      al, not DPclkrst
    out      (dx),al            ;TCK sets low
    or       al,DPclkrst
    out      (dx),al            ;Quiescent again
    ret
pulseclock   ENDP

;====================================================================

getinp       PROC near          ;Get input lines to AL
    inc      dx                 ;Select status port
    in       al,(dx)            ;Got status
    dec      dx                 ;Restore base address
    xor      al,80H             ;Correct polarity of DONE
    ret                         ;AL[6] = TDO, AL[7] = DONE
getinp       ENDP

    END
