.title I2C Bus Driving Routines
.comment $

;=======================================================================;
;                                                                       ;
;   D-X Designs Pty. Ltd.   Perth, Western Australia                    ;
;                                                                       ;
;  Copyright (C) 1997  David R. Brooks                                  ;
;                                                                       ;
;  This program is free software; you can redistribute it and/or        ;
;  modify it under the terms of the GNU General Public License          ;
;  as published by the Free Software Foundation; either version 2       ;
;  of the License, or (at your option) any later version.               ;
;                                                                       ;
;  This program is distributed in the hope that it will be useful,      ;
;  but WITHOUT ANY WARRANTY; without even the implied warranty of       ;
;  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        ;
;  GNU General Public License for more details.                         ;
;                                                                       ;
;  You should have received a copy of the GNU General Public License    ;
;  along with this program; if not, write to the Free Software          ;
;  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.            ;
;                                                                       ;
;=======================================================================;

 Target CPU: Motorola MC68705C8
 Crystal:    4MHz
 Assembler:  2500AD 6805 Macro Assembler - Version 4.05b
$

.pw 120
.pl 88
.symbols on
.spaces on
.asclist off
.condlist off
.linklist

.PAGE0     ;Define basic hardware items

portB:  equ     $01
ddrB:   equ     $05

;---------- I2C control pins (in Port B) -----------

sda:    equ     1               ;SDA bit
scl:    equ     2               ;SCL bit

E2BASE: equ     $A0     ;Base for E2ROM part

.list on
.comment $
 These routines assume the 6805 is always the bus master.

 The I2C lines are open-drain. This is faked by setting
the data-output bits to zero, and driving data (inverted)
to the DDR (direction) bits. An input bit floats high,
while an output bit drives low.

 CAUTION: There is an ugly side-effect in the BSET/BCLR
instructions when used in this way. The detailed action is
Read - Setbit - Write. All 8 bits are re-written, not just
the bit changed. So the output-latches for any bits defined
as Inputs will be set to the current value of the pin,
rather than being retained. In this application, we always
intend the latches to hold zero, so just re-clear them.
This doesn't cause problems by glitching the lines, since
it only happens with a pin set as Input, so the fiddling
of the output latch isn't seen at the pin.
 Of course, since we only twiddle the DDR bits, the data
register shouldn't change. However, it does! Not clear if
this is a problem with the CPU, or just the ICE. But play it
safe.
$

;-------------- External Interface -----------------

.global I2INIT          ;Constructor: start up the I2C bus

                ;Routines for ST24C04 E2ROM (512 bytes)
.global E2READ1         ;E2ROM: read 1 byte
.global E2READN         ;E2ROM: read N bytes
.global E2WRIT1         ;E2ROM: write 1 byte
.global E2WRIT8         ;E2ROM: write 8 bytes

;---------- I2C Macros ------------

sdaget: macro           ;CY := SDA
        brclr   sda,portB,?sdgl#  ;Side effect: bit => CY
?sdgl#:
        endm

sdaput: macro           ;SDA := CY
        bcs     ?sdas#
        bsr     sdaclr
        bra     ?sdax#
?sdas#:
        bsr     sdaset
?sdax#:
        endm

;----------- I2C subroutines -----------

.PAGE0
                ;Temporary variables
i2c1:   db      0       ;Temp. data: getbyte, putbyte
tmp:    db      0       ;Used by WriteAddr
tmp1:   db      0       ;Used by top-level routines

.Program

sdaset:                 ;SDA := 1
        bclr    sda,ddrB        ;Input (float to 1)
        bclr    sda,portB       ;Be sure of these
        bclr    scl,portB
        rts

sdaclr:                 ;SDA := 0
        bset    sda,ddrB        ;Output (drive to 0)
        bclr    sda,portB       ;Be sure of these
        bclr    scl,portB
        rts

sclset:                 ;SCL := 1
        bclr    scl,ddrB
        bclr    sda,portB       ;Be sure of these
        bclr    scl,portB
        rts

sclclr:                 ;SCL := 0
        bset    scl,ddrB
        bclr    sda,portB       ;Be sure of these
        bclr    scl,portB
        rts

sclclk:                 ;Pulse the SCL line
        bsr     sclset
        bsr     sclclr
        rts

stop:                   ;Global STOP signal
        bsr     sdaclr
        bsr     sclset
        bsr     sdaset          ;SDA rises while SCL high
        rts

start:                  ;Global START signal
        bsr     sclset
        bsr     sdaset
        bsr     sdaclr          ;SDA falls while SCL high
        bsr     sclclr
        bsr     sdaset
        rts

I2INIT:                 ;Initialise the I2C bus
        bsr     sdaset          ;Initialise DDR bits
        bsr     sclclr
        bclr    sda,portB       ;Be sure OP bits are low
        bclr    scl,portB
        bsr     stop
        bsr     stop
        bsr     stop
        rts

putbyte:                ;A => I2C bus
        clr     i2c1            ;Bit counter
?1:     rola                    ;A[7] => CY
        sdaput                  ;Set SDA from it
        bsr     sclclk          ;Clock it
        inc     i2c1
        brclr   3,i2c1,?1       ;Loop 8 times
        bsr     sdaset          ;Leave high for ACK
        rts

getbyte:                ;I2C byte => A
        clr     i2c1            ;Bit counter
        bsr     sdaset          ;Be sure it can float
?1:     bsr     sclset          ;Clock UP
        sdaget                  ;SDA => CY
        rola                    ;Shift CY into A[0]
        bsr     sclclr          ;Clock DOWN
        inc     i2c1
        brclr   3,i2c1,?1       ;Loop 8 times
        rts

giveack:                ;If CY=0, ACK: if CY=1, NACK
        sdaput                  ;SDA as reqd.
        bsr     sclclk          ;Clocked through
        bsr     sdaset
        rts

getack:                 ;Get ACK from slave: time out
        bsr     sdaset
        bsr     sclset
        clr     i2c1            ;Time-out counter
?1:     sdaget                  ;ACK?
        bcc     ?2              ;0 - OK
        inc     i2c1            ;1 - wait
        bcc     ?1
?2:     bsr     sclclr
        rts                     ;Done: CY=1 if timed out

logon:                  ;Connect to Slave [A]
        bsr     start
        bsr     putbyte         ;Send the device ID
        bsr     getack          ;Check it's there
        rts                     ;Done: CY=1 if failed

putaddr:                ;A:X = device : address
        bsr     logon           ;Device ID
        bcc     ?1
        rts                     ;Failed to log-on
?1:     txa
        bsr     putbyte         ;Address
        bsr     getack
        rts                     ;Back: CY for error

;------ Routines for ST24C04 E2ROM (512 bytes) -------

.comment $
 The E2ROM routines are written for the SGS 24C04 part.
This is a true E2ROM, having a long write time, and a
paged organisation. It can accept blocks of 8 bytes
(page aligned), and write in parallel. Since the primary
use of block-writes will be to set up an entire array, this
is no problem. FRAM devices have no restrictions, and will
not see the difference.
 E2ROMs indicate "busy writing" by disconnecting from the
bus. Hence a subsequent access will see no ACK to the select,
and will abort. It is for higher-level routines to handle
this event (eg by a wait loop).
$

ReadAddr:       ;Initialise E2ROM chip to read
                        ;CY:X = E2ROM address (9 bits)
                        ; A trashed
        bsr     WriteAddr       ;As a Write, initially
        bcs     ?1              ; - failed
        lda     tmp             ;Get the command
        ora     #1              ;Make it a Read
        bsr     logon           ;Now start the Read
?1:     rts                     ;Done: CY=1 if error


WriteAddr:      ;Initialise E2ROM chip to write
                        ;CY:X = E2ROM address (9 bit)
                        ; A trashed
        lda     #E2BASE         ;Chip base address
        bcc     ?1
        ora     #2              ;CY => A[1] (Addr. MSB)
?1:     sta     tmp             ;Save cmd. for ReadAddr
        bsr     putaddr         ;Set to write
        rts                     ;CY=1 if error


E2READ1:        ;Read 1 byte from the E2ROM to A
                        ;CY:X = E2ROM address (9 bit)
        bsr     ReadAddr        ;Start up
        bcs     ?1              ; - failed
        bsr     getbyte         ;Read 1 byte
        sec
        bsr     giveack         ;Send NACK - the last
        clc
?1:     jsr     stop            ;Close-down the chip
        rts                     ;Data in A, CY=0 if OK


E2READN:        ;Read N bytes from the E2ROM into RAM
                        ;CY:X    = E2ROM address (9 bit)
                        ;A       = RAM address (Buff)
                        ;Buff[0] = Length
        sta     tmp1            ;Save it
        bsr     ReadAddr        ;Set up the chip

getblk:                 ;Common code for block reads
        bcs     ?1              ; - fail
        ldx     tmp1            ;X => buffer
        lda     ,x
        sta     tmp1            ;TMP1 = Count
        bra     ?2

?3:     clc                     ;More to go: ACK
        bsr     giveack
?2:     bsr     getbyte         ;Get 1st/next byte
        sta     ,x
        inx                     ;Move to buffer
        dec     tmp1
        bne     ?3              ;Loop while more
        sec
        bsr     giveack         ;NACK at the end
        clc
?1:     jsr     stop            ;Close down
        rts                     ;Done: CY=0 if OK


E2WRIT1:        ;Write 1 byte from A to E2ROM
                        ;CY:X = E2ROM address (9 bit)
                        ;A    = data
        sta     tmp1
        bsr     WriteAddr       ;Set up to write
        bcs     ?1
        lda     tmp1
        jsr     putbyte         ;Send the byte
        bsr     getack          ;ACK from chip
?1:     jsr     stop            ;Starts the Write
        rts                     ;CY=1 if error


E2WRIT8:        ;Write 8 bytes (1 page) to E2ROM
                        ;CY:X = E2ROM address (9 bit)
                        ;A    = RAM address
        sta     tmp1
        txa
        and     #$7             ;Check for page start
        bne     ?3              ;Fail if not aligned
        bsr     WriteAddr       ;Set up to write
        bcs     ?1
        ldx     tmp1            ;X => buffer
        lda     #8
        sta     tmp1            ;Counter
?2:     lda     ,x
        inx                     ;Next from buffer
        jsr     putbyte         ;Send it
        jsr     getack          ;ACK from chip
        bcs     ?1
        dec     tmp1            ;Count
        bne     ?2              ;Loop
?1:     jsr     stop            ;Done: start the Write
        rts                     ;CY=1 if error

?3:     sec                     ;Failed
        rts

.end
