// Support functions for Intel 28F010 flash PROMs

// D-X Designs, November 1995   D. R. Brooks

#include <stdio.h>
#include <string.h>
#include <mem.h>
#include "dxstds.h"
#include "datafile.h"
#include "prombase.h"
#include "intel.h"
#include "hardware.h"
#include "loader.h"
#include "timing.h"
#include "bargraph.h"

// class Intel {

#define IgetID   0x90       	//Intel commands: part ID
#define Iread    0          	//                reset
#define Ierase   0x20           //                erase setup / erase
#define Ieverf   0xa0  		//                erase verify
#define Iprogm   0x40		//                program setup / program
#define Ipverf   0xc0		//                program verify
#define Ireset   0xff		//                reset

//   public:
Intel::Intel(char *name) {              //Constructor: set top address
   Part.IDbyte[0] = 0x89;          //Identifies Intel
   if (strcmp(name, "28F256A") == 0) {      //28F256A
      MaxAddr = 0x8000;
      Part.IDbyte[1] = 0xb9;
      }
   else {                                   //28F010
      MaxAddr = 0x20000;
      Part.IDbyte[1] = 0xb4;
      }
   }

boolean GetAns(const char *prompt) ;    //User's choice: Yes/No (LOADER.CPP)


void Intel::Identify (chipID *tstID,    //Test for part type
		      chipID *oldID,
		      chipID *newID) {

#define IDlocn 0                        //Part ID address in PROM

      memcpy(tstID,&Part,sizeof(Part));
      ReadBlock (oldID, sizeof(chipID), IDlocn);
      WriteByte (IgetID);              //Activate intelligent ID
      ReadBlock (newID, sizeof(chipID), IDlocn);
      WriteByte (Iread);               //Normal operation
      return;
      };

//======================================================================

   boolean Intel::WritePROM(DataFile *bufptr,
			    long basad) {

      fprintf(stdout,"Checking Intel PROM contents\n");
/*
  Intel allows new data to be written at will, over erased (FF) locns.
  If any locns. to be programmed do not contain FF, the entire chip must
    be erased. This requires first, that all locns. be programmed with
    00, then a global erase command is issued.
*/
      DataFile *testcopy = new DataFile();
      testcopy->BaseAddr = bufptr->BaseAddr;
      testcopy->Length   = bufptr->Length;
      ReadPROM (testcopy,basad);    //See what's there already
      boolean Erase = FALSE;
      boolean Program = FALSE;
      for (word ptr=0; ptr<testcopy->Length; ptr++) {
	 if (testcopy->FileData[ptr] == bufptr->FileData[ptr]) continue;
	 Program = TRUE;
	 if (testcopy->FileData[ptr] != 0xFF) {
	    Erase = TRUE;
	    break;
	    }
	 }
      if (Program == FALSE) {
	 fprintf(stdout,"Already correctly programmed\n");
	 return TRUE;
	 }
      if (Erase == TRUE) {
	 if (GetAns("Confirm erase entire PROM ") == FALSE) return FALSE;
	 QuickErase();
	 }
      fprintf(stdout,"Programming...\n");
      BarGraph cmf(bufptr->Length);
      long ADDR = basad;
      for (word bpt=0; bpt < bufptr->Length; bpt++) {
	 if (QuickPulse(ADDR,bufptr->FileData[bpt]) == FALSE)
	    return FALSE;
	 ADDR++;
         cmf.StepBarGraph(1);
	 }
      WriteByte(Iread);              //End by resetting to "read" mode
      return TRUE;
      };

//======================================================================

//private:

   boolean Intel::QuickPulse(              //Intel programming algorithm
			     long locn,    // One byte only (don't
			     byte data) {  // bump address)
      for (int pulsect = 0; pulsect < 25; pulsect++) {
	 SetAddress(locn);                 //Define the address
	 WriteByte(Iprogm);                //Set-up Program command
	 WriteByte(data);                  //Program data
				   //10uS delay (inherent)
	 WriteByte(Ipverf);                //Program verify
				   // 6uS delay (inherent)
	 if (ReadByte() == data) return TRUE;
	 }                                 //Up to 25 attempts
      return FALSE;                        //Failed to program
      };

//=====================================================================

   boolean Intel::QuickErase() {           //Intel erase algorithm
      fprintf(stdout,"Erasing PROM: zero-ing\n");
      BarGraph comfort(MaxAddr);
      SetAddress(0);
      for (long ADDR=0; ADDR<MaxAddr; ADDR++) {
	 if (ReadByte() != 0) {
	    if (QuickPulse(ADDR,0) == FALSE) return FALSE;
	    }
	 BumpAddr();                 //First, set all to zeros
	 comfort.StepBarGraph(1);
	 }
      comfort.~BarGraph();
      ADDR = 0;
      fprintf(stdout,"Erasing...\n");
      for (int PLSCNT=0; PLSCNT<1000; PLSCNT++) {
	 WriteByte(Ierase);
	 WriteByte(Ierase);	//Initiate the Erase
	 Wait10ms();
	 fprintf(stdout,"Attempt No. %i\n",PLSCNT);
	 BarGraph cmf(MaxAddr - ADDR);
	 while (ADDR < MaxAddr) {
	    SetAddress(ADDR);
	    WriteByte(Ieverf);    //Erase verify
					//Delay 6uS - in software
	    if (ReadByte() != 0xFF) break;
	    if (ADDR == (MaxAddr-1)) {
	       WriteByte(Iread);
	       return TRUE;
	       }
	    ADDR++;
	    cmf.StepBarGraph(1);
	    }
         cmf.~BarGraph();
	 }
      return FALSE;
      };


// } ;          End of class Intel
