Program Disassembler_for_Z80_Programs;
{ Code reverse-assembler for Z-80 source code. Operates similarly to
  the ASMGEN public-domain utility for the 8086.

  D. R. Brooks				June 1990					Perth, Western Australia

  The following .SEQ records are defined:-
  nnnn  B       Byte data
  nnnn  C       Machine code
  nnnn  I       Initial location (default =100H)
  nnnn  S       Text String
  nnnn  W       Word data (Location references)
  nnnn  X       Repeating structure:-
    nn  B                 Count of Bytes
    nn  C                 Count of Code bytes
    nn  S                 Count of String bytes
    nn  W                 Count of Word items

  The source-file (named *.MAC) is suitable for assembly with M80. It
    comprises 4 sections (any of which may be empty):-

  1. Equates defining immediate byte values ------------ labelled inn
  2. Equates defining offsets from the Index Registers - labelled rnn
       NB These are defined as SIGNED DECIMAL values.
  3. Equates defining memory (word) references --------- labelled lnnnn
  4. The code dis-assembly

 N.B. The Z-80 opcode definitions file, OPCODES.Z80, must be in the
      SAME directory as this program's .EXE file.
  }

uses dos;

type
   TextLine  = string[128];

   HexVal    = string[2];

   WildType  = (None,             {Wildcard byte types}
               Constant,
               ByteImmediate,
               CodeRelative,
               IndexOffset,
               WordLSB,
               WordMSB );

   CodeEntry = record
               InstrSize : byte;                    {No. of instruction bytes}
               Code      : array[1..4] of byte;     {Constant code bytes}
               Wild      : array[1..4] of WildType; {Wildcard byte types}
               Source    : string[20];              {Equivalent source code}
               EndBlock  : boolean;                 {Uncond. jump - newline}
               end;

   VirtualPage = array[0..16383] of byte;   {16kB virtual store page}
   PagePointer = ^VirtualPage;              {Access pointer}
   MemBlock    = array[0..3] of PagePointer;

   SeqType     = (InitialLocation,          {Sequence file record types}
                  Terminator,
                  ByteData,
                  MachineCode,
                  StringData,
                  WordData,
                  Repeating );

   SequenceEntry = record                   {Sequence-control file}
                   SeqLoc  : word;
                   SeqMode : SeqType;
                   end;
var
   RefTable,                                       {Reference table}
   VirtualStore : MemBlock;                        {Virtual store}
   CodeTable    : array[1..1023] of CodeEntry;     {Disassembly table}
   SequenceTable: array[0..1023] of SequenceEntry; {Sequence Data}
   CurrentSeq,
   MaxSequence,
   MaxSourceLine: integer;
   CurrentAddress,
   InitialAddress,               {File load address}
   FinalAddress   : word;        {End of code load}

   ThisByteImmediate : byte;     {Values obtained by current instruction}
   ThisCodeRelative  : word;
   ThisIndexOffset   : byte;
   ThisWord          : word;

   CodeDir    : DirStr;          {File base-name, as input}
   CodeName   : NameStr;

   outfile    : text;            {Source-code output file}

const
   WordReference  : byte = 1;    {Bit-mask values in Reference table}
   ByteReference  : byte = 2;
   IndexReference : byte = 4;


   MaxColumn : integer = 50;           {Limit for operand part of line}



function HexIn (ASCII  :TextLine;    {Source text}
                Locn   :integer;     {Locn. to read}
            var Answer :byte )       {Result}
                       :boolean;     {Convert 1 byte ASCII to Binary}
var
   result : array[0..1] of byte;
   ctr    : integer;
   okay   : boolean;
   value  : char;

begin
   okay:= true;
   for ctr:= 0 to 1 do begin
      value:= upcase(ASCII[Locn+ctr]);
      case value of
         '0'..'9': result[ctr]:= ord(value) - ord('0');
         'A'..'F': result[ctr]:= ord(value) + 10 - ord('A');
         else      okay:= false;
         end
      end;
   HexIn:= okay;
   if okay then Answer:= (result[0] shl 4) + result[1];
   end;


function HexOut(item :byte) :HexVal;    {Output in hexadecimal}
const
   HexChar : array[0..15] of char =
             ('0', '1', '2', '3', '4', '5', '6', '7',
              '8', '9', 'a', 'b', 'c', 'd', 'e', 'f' );

begin
   HexOut:= '  ';                       {Length:= 2}
   HexOut[1]:= HexChar[(item shr 4) and 15];
   HexOut[2]:= HexChar[item and 15];
   end;



                                        {Leading-zero suppression}
function HexValue(n :word; var size :integer) : TextLine;
var
   local : string[8];
begin
   local:= '0' + HexOut(hi(n)) + HexOut(lo(n));
   while ((length(local) > 1) and
          (local[1] = '0') and
          (local[2] <= '9')) do delete(local,1,1);   {Strip leading zeros}
   if local[length(local)] in ['b','d'] then local:= local + 'h';
   size:= length(local);
   HexValue:= local;
   end;


function GetCode(Offset :word) :byte;
begin                                   {Get byte from virtual storage}
   GetCode:= VirtualStore[(Offset shr 14) and 3]^[Offset and $3fff];
   end;


procedure initialise;
var
   SourceFile : text;
   SourceLine : TextLine;
   BinaryFile : file;
   count1,
   count2,
   count3     : integer;
   wpt1       : longint;
   wpt2       : word;
   stub       : string[2];
   CodeExtn   : ExtStr;
   SeqAddr    : TextLine;
   Seqlsb,
   Seqmsb     : byte;
   LoadBlock  : array[0..1023] of byte;

const
   WhiteSpace : set of char = [^I,' '];


   procedure Fail(echo :boolean; ermesg :TextLine);
   begin
      if echo then writeln(SourceLine);
      writeln(ermesg);
      halt;
      end;

function ProgramPath : DirStr;           {Get path to this program}
type
   env = array[0..32767] of char;

var
   eptr   : ^env;
   i      : integer;
   work   : PathStr;
   direct : DirStr;
   name   : NameStr;
   extn   : ExtStr;
   dosreg : Registers;

   function dosok : boolean;           {Needs DOS-3 for this technique}
   begin
      with dosreg do begin
         AH:= $30;                     {Get DOS version}
         MsDos(dosreg);
         dosok:= AL >= 3;              {Must be DOS 3.xx or later}
         end
      end;


begin
   if dosok then begin
      with dosreg do begin
         AH:= $62;                     {Get PSP segment}
         MsDos(dosreg);
         eptr:= Ptr(MemW[BX:$2C],0);   {Point at Envt. segment}
         end;
      i:= 0;
      repeat                           {Find end of environment strings}
         inc(i);
         until ((eptr^[i] =#0) and (eptr^[i+1] =#0));
      i:= i+4;                         {Move forward to program path}
      work:= '';
      while eptr^[i] <> #0 do begin
         work:= work + eptr^[i];       {Copy over the string}
         inc(i);
         end;
      FSplit(work, direct, name, extn);
      ProgramPath:= direct;
      end
   else
      Fail(false, 'DOS 3.xx is required');
   end;


begin
   InitialAddress:= $100;           {CP/M default address}
   writeln('Opening input files');
   FSplit(paramstr(1), CodeDir, CodeName, CodeExtn);
   assign(SourceFile, CodeDir + CodeName + '.SEQ');
   reset(SourceFile);
   count1:= 0;
   while not eof(SourceFile) do begin
      with SequenceTable[count1] do begin
         readln(SourceFile,SourceLine);
         count2:= 1;
         while SourceLine[count2] in WhiteSpace do count2:= count2+1;
         SeqAddr:= '';
         while not (SourceLine[count2] in WhiteSpace) do begin
            SeqAddr:= SeqAddr + SourceLine[count2];
            inc(count2);
            end;
         if length(SeqAddr)>4 then delete(SeqAddr, 1, length(SeqAddr)-4);
         while length(SeqAddr)<4 do SeqAddr:= '0' + SeqAddr;
         if not ( HexIn(SeqAddr, 1, Seqmsb) and
                  Hexin(Seqaddr, 3, Seqlsb)) then begin
                  Fail(true,'Bad Sequence-file address');
                  end;
         Seqloc:= ((Seqmsb shl 8) + Seqlsb) and $ffff;
         while SourceLine[count2] in WhiteSpace do inc(count2);
         case upcase(SourceLine[count2]) of
            'B': SeqMode:= ByteData;
            'C': SeqMode:= MachineCode;
            'I': begin
                 InitialAddress:= SeqLoc;
                 SeqMode:= InitialLocation;
                 end;
            'S': SeqMode:= StringData;
            'W': SeqMode:= WordData;
            'X': SeqMode:= Repeating;
            else begin
                 Fail(true,'Invalid control letter');
                 end
            end
         end;
      inc(count1);
      end;
   close(SourceFile);
   MaxSequence:= count1-1;

   writeln('Allocating memory space');
   for count1:= 0 to 3 do begin
      new(VirtualStore[count1]);
      FillChar(VirtualStore[count1]^[0], 16384, 0);
      new(RefTable[count1]);
      FillChar(RefTable[count1]^[0], 16384, 0);
      end;

   writeln('Loading binary file - <', paramstr(1), '>');
   assign(BinaryFile, paramstr(1));
   reset(BinaryFile,1);            {Input file - untyped bytes}
   wpt1:= FileSize(BinaryFile);
   if (wpt1 + InitialAddress) > 65535 then
      Fail(false,'File overflows 65kB');
   wpt2:= InitialAddress;
   while not eof(BinaryFile) do begin
      BlockRead(BinaryFile, LoadBlock, 1024, count3 );
      for count1:= 0 to count3-1 do begin
         VirtualStore[(wpt2 shr 14) and 3]^[wpt2 and $3fff]:=
                      LoadBlock[count1];
         inc(wpt2);
         end
      end;
   close(BinaryFile);
   FinalAddress:= InitialAddress + wpt1;

   writeln('Loading source-mnemonic file');
   assign(SourceFile, ProgramPath + 'OPCODES.Z80');
   reset(SourceFile);              {Load up the source-code data}
   count1:= 1;
   while not eof(SourceFile) do begin
      readln(SourceFile,SourceLine);
      with CodeTable[count1] do begin
         for count3:= 1 to 4 do begin
            stub:= copy(SourceLine, (count3-1)*3 +1, 2);
            Code[count3]:= 0;
            if stub = '  ' then Wild[count3]:= None          else
            if stub = 'nn' then Wild[count3]:= ByteImmediate else
            if stub = 'rr' then Wild[count3]:= CodeRelative  else
            if stub = 'xx' then Wild[count3]:= IndexOffset   else
            if stub = 'yy' then Wild[count3]:= WordMSB       else
            if stub = 'zz' then Wild[count3]:= WordLSB       else begin
               Wild[count3]:= Constant;
               if not (HexIn(stub, 1, Code[count3])) then
                  Fail(true, 'Invalid Code-byte type declarator');
               end;
            if stub <> '  ' then InstrSize:= count3;
            end;
         Source:= copy(SourceLine, 24, length(SourceLine)+1);
         EndBlock:= (SourceLine[20] = 'j');
         end;
      count1:= count1 +1;
      end;
   MaxSourceLine:= count1-1;
   close(SourceFile);
   end;

procedure expandsequence;          {Expands repeat-structures in SEQ file}
var
   tempseq   : array[0..1023] of SequenceEntry;
   repbase,
   x, y, z   : integer;
   curloc    : word;

begin                              {First, make local copy}
   writeln('Expanding repeat-sequences');
   for x:= 0 to MaxSequence do tempseq[x]:= SequenceTable[x];
   tempseq[MaxSequence+1].SeqLoc:= FinalAddress+1;  {Dummy top entry}
   tempseq[MaxSequence+1].SeqMode:= Terminator;
   inc(MaxSequence);
   y:= 0;                          {New place-pointer}
   x:= 0;                          {Source pointer}
   while x<= MaxSequence+1 do begin
      case tempseq[x].SeqMode of
         InitialLocation: begin    {Ignore it - it's been done}
                             x:= x+1;
                             end;
         Repeating      : begin    {Expand repeat blocks}
                             repbase:= x;
                             curloc := tempseq[x].SeqLoc;  {Start of seq.}
                             z:= x+1;
                             while tempseq[z].SeqLoc<tempseq[x].SeqLoc do
                                z:= z+1;                   {End of it}
                             while curloc<tempseq[z].SeqLoc do begin
                                if x<z-1 then inc(x)
                                         else x:= repbase+1;
                                SequenceTable[y].SeqLoc:= curloc;
                                SequenceTable[y].SeqMode:= tempseq[x].SeqMode;
                                curloc:= curloc + tempseq[x].SeqLoc;
                                          {Double-scaling for Word Data}
                                if tempseq[x].SeqMode = WordData then
                                   curloc:= curloc + tempseq[x].SeqLoc;
                                y:= y+1;
                                end;
                             x:= z;
                             end;
         else             begin    {Others - copy across}
                             SequenceTable[y]:= tempseq[x];
                             y:= y+1;
                             x:= x+1;
                             end
         end                       {End the CASE}
      end;
   MaxSequence:= y-1;
   end;


function FindOpcode : integer; {Match instruction at CurrentAddress}
type
  testres   = (low, equal, high);

var
  clocn,
  cdist     : integer;
  result    : testres;
  instr     : array[1..4] of byte;

  function testit : testres;         {The actual comparison}
  var
     tcode,
     tinstr,
     index : integer;
     check : testres;

  begin
     if clocn > MaxSourceLine then testit:= low      {Off end of table}
                              else with CodeTable[clocn] do begin
        index:= 1;
        repeat
           tcode := Code[index] and $ff;
           tinstr:= instr[index] and $ff;
           if Wild[index] <> Constant    then check:= equal else
           if tinstr < tcode             then check:= low   else
           if tinstr > tcode             then check:= high  else
                                              check:= equal;
           index:= index+1;
           until ((index > 4) or (check <> equal));
        testit:= check;
        end
     end;

begin                                {Use binary search vs. opcode table}
   for clocn:= 1 to 4 do instr[clocn]:= GetCode(CurrentAddress+clocn-1);
   clocn := 0;
   cdist := 1024;
   result:= high;
   while cdist > 1 do begin
      cdist:= cdist div 2;
      case result of
         low:   begin
                clocn:= clocn - cdist;
                result:= testit;
                end;
         equal: cdist:= 0;
         high:  begin
                clocn:= clocn + cdist;
                result:= testit;
                end
         end
      end;
   if result = equal then with CodeTable[clocn] do begin
            ThisWord:= 0;
            FindOpcode:= clocn;    {Found: index table}
            for cdist:= 1 to 4 do
               case Wild[cdist] of
                  ByteImmediate: ThisByteImmediate:= instr[cdist];
                  CodeRelative : begin         {Sign-extended offset}
                                    ThisCodeRelative:= instr[cdist] and $ff;
                                    if ThisCodeRelative > $7f then
                                       ThisCodeRelative:= ThisCodeRelative or $ff00;
                                    ThisCodeRelative:= ThisCodeRelative +
                                                       CurrentAddress + 2;
                                    end;
                  IndexOffset  : ThisIndexOffset:= instr[cdist];
                  WordLSB      : ThisWord:= ThisWord or instr[cdist];
                  WordMSB      : ThisWord:= ThisWord or (instr[cdist] shl 8);
                  end    {CASE}
            end
         else FindOpcode:= -1;     {Not found signal}
   end;


procedure makereferencetables;       {Locate References}
var
   x,
   instptr   : integer;

   procedure MakeReference(offset :word; reftype :byte);
   begin                             {Logical OR with byte in virtual storage}
      RefTable[(Offset shr 14) and 3]^[Offset and $3fff]:=
         RefTable[(Offset shr 14) and 3]^[Offset and $3fff] or reftype;
      end;

   procedure wordref(locn :word); {Indirect WORD reference}
   var
      refs : word;

   begin
      refs:= ((GetCode(locn+1) shl 8) and $ff00) + GetCode(locn);
      MakeReference(refs, WordReference);
      end;

begin
   writeln('Building reference tables');
   CurrentAddress:= SequenceTable[0].SeqLoc;
   for CurrentSeq:= 0 to MaxSequence-1 do begin
      case SequenceTable[CurrentSeq].SeqMode of
         WordData:    begin           {Word References}
                      while CurrentAddress < SequenceTable[CurrentSeq+1].SeqLoc do begin
                         wordref(CurrentAddress);
                         CurrentAddress:= CurrentAddress+2;
                         end
                      end;
         MachineCode: begin           {Code References}
                      while CurrentAddress < SequenceTable[CurrentSeq+1].SeqLoc do begin
                         instptr:= FindOpcode; {Try to match code}
                         if instptr < 0 then inc(CurrentAddress) {Bad code: skip byte}
                                        else with CodeTable[instptr] do begin
                            for x:= 1 to 4 do begin
                               case Wild[x] of
                                  ByteImmediate: begin
                                                 MakeReference (
                                                    (ThisByteImmediate and $ff),
                                                    ByteReference );
                                                 end;
                                  CodeRelative : begin
                                                 MakeReference (
                                                    ThisCodeRelative,
                                                    WordReference );
                                                 end;
                                  IndexOffset  : begin
                                                 MakeReference (
                                                    (ThisIndexOffset and $ff),
                                                    IndexReference );
                                                 end;
                                  WordLSB      : begin
                                                 MakeReference (
                                                    ThisWord,
                                                    WordReference );
                                                 end;
                                  end
                               end;
                            CurrentAddress:= CurrentAddress + InstrSize;
                            end
                         end
                      end;
                                      {Other Sequence items do not generate refs.}
                 else CurrentAddress:= SequenceTable[CurrentSeq+1].SeqLoc;
         end                          {End the CASE}
      end
   end;


procedure OutputByteReferences;
var
   dummy,
   count1 : integer;

begin
   for count1:= 0 to 255 do            {Limits of Byte values}
      if (ByteReference and RefTable[0]^[count1]) <> 0 then
         writeln(outfile, 'i', HexOut(count1 and $ff), ^I,
                          'equ', ^I, HexValue((count1 and $ff), dummy));
      end;


procedure OutputIndexReferences;        {NB These are signed offsets}
var
   count1 : integer;

begin
   for count1:= 0 to 255 do             {Limits of Index offsets}
      if (IndexReference and RefTable[0]^[count1]) <> 0 then begin
         write(outfile, 'r', HexOut(count1 and $ff), ^I, 'equ', ^I);
         if count1 > 127 then writeln(outfile, 256-count1)
                         else writeln(outfile, count1);
         end
   end;


function Referenced(refloc :word) :boolean;  {Test & reset WORD refs.}
begin
   Referenced:= false;
   if (RefTable[(refloc shr 14) and 3]^[refloc and $3fff] and
      WordReference) <> 0 then begin
      Referenced:= true;
      RefTable[(refloc shr 14) and 3]^[refloc and $3fff]:=
         RefTable[(refloc shr 14) and 3]^[refloc and $3fff] and
            (not WordReference);
      end
   end;



procedure OutputAddressReferences;
var
   count1 : word;
   dummy  : integer;

begin
   for count1:= 0 to 65535 do
      if ((count1 < InitialAddress) or
          (count1 > FinalAddress)) then begin
          if Referenced(count1) then
             writeln(outfile, 'l', HexOut(hi(count1)), HexOut(lo(count1)),
                              ^I, 'equ', ^I, HexValue(count1, dummy));
          end
   end;



procedure writereferencetables;
begin
   writeln('Writing reference tables');
   writeln(outfile,'.radix', ^I, '16');
   writeln(outfile);
   writeln(outfile,^I, 'dseg', ^I, ^I, ^I, ';Byte References');
   OutputByteReferences;
   writeln(outfile);
   writeln(outfile,'.radix', ^I, '10', ^I, ^I, ^I, ';Index Offsets');
   OutputIndexReferences;
   writeln(outfile);
   writeln(outfile,'.radix', ^I, '16', ^I, ^I, ^I, ';Address References');
   OutputAddressReferences;
   end;




procedure FixColumn                    {New output line if too long}
               ( var column : integer;      {Current column count}
                       tail : char );       {Append if creating new line}

begin
   if column <> 0 then begin
      writeln(outfile,tail);
      column:= 0;
      end
   end;



procedure References                   {Insert address references in output}
                   (   back : word;         {No. of bytes to look backward}
                 var column : integer;      {Output-line column count}
                       tail : char );       {Append if creating new line}
var
   n : integer;

begin
   for n:= back downto 0 do begin
      if ((CurrentAddress >= n) and (Referenced(CurrentAddress-n))) then begin
         FixColumn(column, tail);
         tail:= ' ';
         write(outfile, 'l', HexOut(hi(CurrentAddress-n)),
                             HexOut(lo(CurrentAddress-n)) );
         if n=0 then write(outfile, ':')
                else writeln(outfile, ^I, 'equ', ^I, '$-', n);
         end
      end
   end;



procedure OutputBytes (Top : word);    {Output BYTE data}
var
   back : word;
   temp,
   coln : integer;

begin
   coln:= 0;
   back:= 4;                           {Look back 4 bytes (instr. before?) }
   while CurrentAddress < Top do begin
      References(back, coln, ' ');
      back:= 0;                        {No need to look back, now}
      if coln=0 then write(outfile, ^I, 'db', ^I)
                else write(outfile, ',');
      write(outfile, HexValue(GetCode(CurrentAddress), temp) );
      inc(CurrentAddress);
      coln:= coln + temp + 1;          {Space written out}
      if coln > MaxColumn then FixColumn(coln, ' ');
      end;
   FixColumn(coln, ' ');
   end;


procedure OutputCode (Top : word);     {Output CODE data}
var
   back : word;
   n,
   coln : integer;
   work : string[30];


   procedure PutOut(x :byte);          {Merge byte value into output}
   var
      y : HexVal;
   begin
      y:= HexOut(x);
      work[n]  := y[1];
      work[n+1]:= y[2];
      end;

begin
   back:= 4;
   while CurrentAddress < Top do begin
      coln:= 0;
      References(back, coln, ' ');
      coln:= FindOpcode;               {coln now used for opcode ptr.}
      if coln<0 then begin                   {Illegal opcode - output byte}
         writeln(outfile, ^I, 'db', ^I,
                          HexValue((GetCode(CurrentAddress) and $ff), n) );
         back:= 0;
         inc(CurrentAddress);
         end
      else with CodeTable[coln] do begin
         work:= ^I + Source;
         back:= InstrSize;
         CurrentAddress:= CurrentAddress + InstrSize;
         for n:= 2 to length(work)-1 do
            case work[n] of
               'N': PutOut(ThisByteImmediate);
               'X': PutOut(ThisIndexOffset);
               'Y': PutOut(hi(ThisWord));
               'Z': PutOut(lo(ThisWord));
               'P': PutOut(hi(ThisCodeRelative));
               'Q': PutOut(lo(ThisCodeRelative));
               end;
         writeln(outfile, work);
         if EndBlock then writeln(outfile);   {Extra line after u/c jump}
         end
      end;
   end;



procedure OutputStrings (Top : word);  {Output STRING data}
var
   back : word;
   temp,
   coln : integer;
   tail : char;

label
   BeginLine,
   DoLimit,
   NotString;


   function printable : boolean;
   begin
      printable:= chr(GetCode(CurrentAddress)) in
                  [' '..'&', '('..'~'];
      end;

begin
   tail:= ' ';
   back:= 4;
   coln:= 0;
   while CurrentAddress < Top do begin
      References(back, coln, tail);
      back:= 0;                        {1 byte at a time}
      if coln=0 then begin
         tail:= ' ';
         write(outfile, ^I, 'db', ^I); {Begin new line}
         goto BeginLine;
         end
      else begin
         if tail = ' ' then goto NotString
                       else begin   {We are presently outputting a string...}
            if printable then begin    {Just extend the present string}
               write(outfile, chr(GetCode(CurrentAddress)) );
               inc(coln);
               goto DoLimit;
               end
            else begin                 {Unprintable - end the current string}
               write(outfile, '''');
               tail:= ' ';
               inc(coln);
               goto NotString;
               end
            end
         end;
NotString:
      write(outfile, ',');           {Unprintable - list of bytes}
      inc(coln);
BeginLine:                           {Start, or next byte}
      if printable then begin
         write(outfile, '''', chr(GetCode(CurrentAddress)) );
         tail:= '''';                {Switch into string mode}
         temp:= 2;
         end
      else write(outfile, HexValue((GetCode(CurrentAddress) and $ff), temp));
      coln:= coln+temp;
DoLimit:
      if coln > MaxColumn then FixColumn(coln, tail);
      inc(CurrentAddress);
      end;
   FixColumn(coln, tail);
   end;



procedure OutputWords (Top : word);    {Output WORD data}
var
   back : word;
   coln : integer;

begin
   coln:= 0;
   back:= 4;
   while CurrentAddress < Top do begin
     References(back, coln, ' ');
     back:= 1;                         {Words: look backward 1 byte}
     if coln=0 then write(outfile, ^I, 'dw', ^I)
               else write(outfile, ',');
     write(outfile, 'l', HexOut(GetCode(CurrentAddress+1)),
                         HexOut(GetCode(CurrentAddress)) );
     CurrentAddress:= CurrentAddress+2;
     coln:= coln+6;                    {Amount written}
     if coln > MaxColumn then FixColumn(coln, ' ');
     end;
   FixColumn(coln, ' ');
   end;



procedure outputsourcecode;
var
   dummy :integer;

begin
   writeln('Writing source-code');
   writeln(outfile);
   writeln(outfile, ^I, 'cseg', ^I, ^I, ^I, ';Source dis-assembly');
   CurrentAddress:= SequenceTable[0].SeqLoc;
   writeln(outfile, ^I, 'org',  ^I, HexValue(CurrentAddress, dummy) );
   for CurrentSeq:= 0 to MaxSequence-1 do begin
      case SequenceTable[CurrentSeq].SeqMode of
         ByteData    : OutputBytes(SequenceTable[CurrentSeq+1].SeqLoc);
         MachineCode : OutputCode(SequenceTable[CurrentSeq+1].SeqLoc);
         StringData  : OutputStrings(SequenceTable[CurrentSeq+1].SeqLoc);
         WordData    : OutputWords(SequenceTable[CurrentSeq+1].SeqLoc);
         end
      end
   end;



begin                     { M A I N L I N E }
   initialise;
   expandsequence;
   makereferencetables;
   assign(outfile, CodeDir + CodeName + '.MAC');
   rewrite(outfile);
   writeln(outfile,'.z80');
   writereferencetables;
   outputsourcecode;
   writeln(outfile,^I,'end');
   close(outfile);
   end.
